/* from xen/arch/x86/include/asm/types.h */

#ifndef _ASM_PPC_TYPES_H
#define _ASM_PPC_TYPES_H

typedef unsigned long paddr_t;
typedef unsigned long vaddr_t;
#define INVALID_PADDR (~0UL)
#define PRIpaddr "016lx"

#endif /* _ASM_PPC_TYPES_H */
