/* SPDX-License-Identifier: GPL-2.0-only */
#ifndef __XEN_HYPERCALL_H__
#error "asm/hypercall.h should not be included directly - include xen/hypercall.h instead"
#endif

#ifndef __ASM_GENERIC_HYPERCALL_H__
#define __ASM_GENERIC_HYPERCALL_H__

#endif /* __ASM_GENERIC_HYPERCALL_H__ */

/*
 * Local variables:
 * mode: C
 * c-file-style: "BSD"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
