/**
 * @file
 * @section AUTHORS
 *
 * Copyright (C) 2021 EPAM Systems Inc.
 *
 * @section LICENSE
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; If not, see <http://www.gnu.org/licenses/>.
 *
 * @section DESCRIPTION
 *
 *  This file contains common libxenvchan declarations.
 */
#ifndef LIBVCHAN_H
#define LIBVCHAN_H

void close_xs_srv(struct libxenvchan *ctrl);

#endif /* LIBVCHAN_H */
